--- Todo: turn this thing into widget handler that will load other lua files for doing different things
--[[
Functions of the engine:

LogMessage(...)
logs the message.
GetOption(string)
gets option, returns string or nothing if option does not exist. Todo: table for options
SetOption(string, value)
Sets the option.


Coding convention:

The functions with_underscope_names are called by engine.
Returning true from such function supresses engine's built-in handler, whenever that makes sense.
For example, sound_on_chaser_death is called by engine, and if you add return true to it, engine will not make sound by itself.
(chaser is the name of those things that run after player and shoot at him)

]]


local debug=false


-- Following parameters are updated by the engine, and are listed here for convenience only
audio_visualizer_in={
amplitude=0,
smooth_amplitude=0,
beat_amplitude=0, -- equals amplitude/smooth_amplitude when smooth amplitude is greater than some threshold
}
-- Following parameters have to be set by scripts
audio_visualizer_out={
background_offset_x=2,-- shifting of background over time
background_offset_y=0,
background_update_speed=0,-- speed of updating of background, combines with frameskip
background_phase={0,0,0} -- use this for blinking. Shifts the colours. format is rgb
}
local avi=audio_visualizer_in --- shorter abbrevations (also, locals like that make script marginally faster)
local avo=audio_visualizer_out
-- By the way. Variables that you create should be local


--[[

local level_script = nil

local level_script_env = nil

-- Current global environment appears read only to the level script,
-- to minimize risk of script leaving some changes as levels are reloaded.
-- todo: in C++ , wipe our entire lua environment on level change?
local eproxy = {}
eproxy.__parent = getfenv(0)
eproxy.__index = function(t, n)
	local parent = rawget(getmetatable(t), "__parent")
	if env and rawget(t,n) then return rawget(env,n) end

	if not parent then return nil end
	return parent[n]
end
eproxy.__newindex = function(t, n, v)
	local parent = rawget(getmetatable(t), "__parent")
	-- if rawget(parent,n) then parent[n] = v end
	rawset(t,n,v)
end

function on_load_level(a)
	LogMessage("lua: on_load_level", a)
	if a and a~="" then
		local filename=VFSTranslate("$$data/levels/" .. a);
		local err
		level_script, err=loadfile(filename)
		if not level_script then
			LogMessage("Lua error loading ", filename, ":", err)
			return
		end
		level_script_env = {}
		setmetatable(level_script_env,eproxy)
		setfenv(level_script,level_script_env)
		local status, msg=pcall(level_script)
		if not status then
			LogMessage("Lua error running ", filename, ":", err)
		end
		for n,v in pairs(level_script_env) do
			LogMessage(n,"=",v)
		end
	else
		level_script_env = nil
		level_script = nil
	end
end
--]]

--- don't draw with opengl from on_ event handlers

function on_game_frame(dt) --- called by engine
	--LogMessage("lua on_game_frame",dt)
end

function on_screen_frame(dt) --- called by engine
	--LogMessage("lua on_screen_frame",dt)
end


local audio_x=0 --- used by drawer
local notes={};
local note_smoothing_time=0.1
local smooth_notes={};
local smooth_avg_amp=0
for i=1,512 do -- hack
		smooth_notes[i]=0
end

function on_audio_frame(dt) --- called by engine
	-- spams the console
	--LogMessage("lua on_audio_frame", audio_visualizer_in.amplitude, audio_visualizer_in.smooth_amplitude, audio_visualizer_in.beat_amplitude)
	local a=math.log(avi.amplitude*1E2 + 1)/math.log(1E2 + 1) --- maps amplitudes 0.01 .. 1 to 0..1 logarithmically
	local phase=a
	--avo.background_phase={phase,phase,phase}
	audio_x=audio_x+3

	notes=GetMusicNotes();
	if notes==nil then return end
	--- notes.steps_per_octave gives number of notes per octave (currently 48) ,
	--- notes.A440 gives index of note A with frequency 440 hz
	--- notes start from A subsubcontra and go up all way to about 22khz .

	local a=math.exp(-dt/note_smoothing_time)
	smooth_avg_amp=0;
	for i=1,#notes do
		smooth_notes[i]=smooth_notes[i]*a+notes[i]*(1-a)
		smooth_avg_amp=smooth_avg_amp+smooth_notes[i]
	end
	smooth_avg_amp=smooth_avg_amp/#notes
end

--- draw with opengl from draw_ functions

function draw_HUD(width,height,bloom)-- called by engine, return true to suppress built-in drawer.
	if bloom then return false end

	if not debug then return false end--- dont want this function now.

	if not notes then return end
	gl.Enable(gl.BLEND)
	gl.Begin(gl.QUADS);
	local piano_keyboard_height=20

	local black={false,true,false,true,false,false,true,false,true,false,true,false}

	local halftones=#notes*12/notes.steps_per_octave
	local halftone_width=notes.steps_per_octave/12

	for i=2,#notes-1 do
		local a=smooth_notes[i]
		local is_peak=a>0.52*(smooth_notes[i-1]+smooth_notes[i+1]) and a>smooth_avg_amp
		local x=i

		if is_peak then
			gl.Color(1,1,0)
		else
			gl.Color(0,1,0)
		end
		a=(math.log(a*1E5))*0.05
		if a>0 then
			--- vertices of a rectangle are specified like this
			gl.Vertex(x+width-#notes, height - piano_keyboard_height)
			gl.Vertex(x+width-#notes +1, height - piano_keyboard_height)
			gl.Vertex(x+width-#notes +1, height-a*height - piano_keyboard_height)
			gl.Vertex(x+width-#notes, height-a*height - piano_keyboard_height)
		end
	end
	--- draw ocrave markers
	local octaves=#notes/notes.steps_per_octave
	for i=0,octaves-1 do
		local x=i*notes.steps_per_octave+3*halftone_width
		gl.Color(0,0.1,0)
		gl.Vertex(x+width-#notes, height)
		gl.Vertex(x+width-#notes, height-300)
		gl.Color(0,0.2,0)
		gl.Vertex(x+width-#notes +notes.steps_per_octave, height - 300)
		gl.Vertex(x+width-#notes +notes.steps_per_octave, height)
	end

	--- draw 'keyboard' underneath
	for i=0,halftones-1 do
		local is_black=black[((i-3)%12) +1]
		if is_black then
			gl.Color(0,0.1,0)
		else
			gl.Color(0,0.3,0)
		end
		local x=i*halftone_width
		gl.Vertex(x+width-#notes, height )
		gl.Vertex(x+width-#notes +halftone_width-1, height)
		gl.Vertex(x+width-#notes +halftone_width-1, height - piano_keyboard_height)
		gl.Vertex(x+width-#notes, height - piano_keyboard_height)
	end

	gl.End();

	return false -- do not override engine's DrawHUD
end

function draw_audio_onto_background(width,height)  --- called by engine, twice per frame now. return true to suppress built-in drawer.
	audio_x=audio_x%width
	--LogMessage("draw_audio_onto_background")
--[[
	gl.Begin(gl.QUADS);
	for i=2,#notes-1 do
		local a=smooth_notes[i]
		local is_peak=a>0.9*(smooth_notes[i-1]+smooth_notes[i+1]) and a>smooth_avg_amp and a<notes[i]
		local x=(i)/#notes

		if is_peak then
			local y0=0.5*(i-1)/#notes
			local y=0.5*(i)/#notes
			gl.Color(0,a,0)
			gl.Vertex((2.0*audio_x)/width-1.0, y0)
			gl.Vertex((2.0*audio_x)/width-1.0, y)

			gl.Color(a,0,0)
			gl.Vertex(2.0*(audio_x+4.0)/width-1.0, y)
			gl.Vertex(2.0*(audio_x+4.0)/width-1.0, y0)

			gl.Color(0,a,0)
			gl.Vertex((2.0*audio_x)/width-1.0, -y0)
			gl.Vertex((2.0*audio_x)/width-1.0, -y)

			gl.Color(a,0,0)
			gl.Vertex(2.0*(audio_x+4.0)/width-1.0, -y)
			gl.Vertex(2.0*(audio_x+4.0)/width-1.0, -y0)
		end
	end
	gl.End();
	]]

	gl.Begin(gl.QUADS);
	local a=avi.amplitude
	gl.Color(0,1,0)
	gl.Vertex((2.0*audio_x)/width-1.0,-a)

	gl.Color(0,1,0)
	gl.Vertex((2.0*audio_x)/width-1.0, -a/2)

	gl.Color(1,0,0)
	gl.Vertex(2.0*(audio_x+4.0)/width-1.0, -a/2)

	gl.Color(1,0,0)
	gl.Vertex(2.0*(audio_x+4.0)/width-1.0,-a)


	gl.Color(0,1,0)
	gl.Vertex((2.0*audio_x)/width-1.0,a)

	gl.Color(0,1,0)
	gl.Vertex((2.0*audio_x)/width-1.0, a/2)

	gl.Color(1,0,0)
	gl.Vertex(2.0*(audio_x+4.0)/width-1.0, a/2)

	gl.Color(1,0,0)
	gl.Vertex(2.0*(audio_x+4.0)/width-1.0,a)
	gl.End();


	return true-- if you return true, built-in drawer does not run.
end

--- px,py.pz is position of impact; vx,vy,vz is velocity of object; bvx,bvy,bvz is velocity of bullet.
function sound_on_chaser_hit(health, px,py,pz, vx,vy,vz, bvx,bvy,bvz)-- called by engine, return true to suppress built-in.

end

function on_chaser_death(px,py,pz, vx,vy,vz)-- called by engine, return true to suppress built-in.

end

local function NoteOffsetToPitch(i)-- i: offset in half tones. returns: the pitch value for given offset. you need to multiply pitch values
	return 2.0^(i/12.0)
end

function sound_on_chaser_death(px,py,pz, vx,vy,vz)-- called by engine, return true to suppress built-in.
	so={pos={px,py,pz}, file="$$data/click1.wav", lifetime=0.5, vel={vx,vy,vz}, sound_gain=1, sound_pitch=0.25+2.0*math.random()}
	LogMessage("creating sound from lua")
	NewSound(so)
	return true
end

LogMessage("main.lua loaded!")

